#include <windows.h>
#include <shlwapi.h>
#include <stdio.h>
#include "../general.h"
#include "pool.h"
#include "../emulate.h"


/* static functions for Internal use only */

static u32 MmPoolPageStart(MEMORY_POOL *Pool) {
	return (sizeof(MEMORY_POOL_HEADER) + CPU_CACHE_LINE_SIZE - 1) & ~(CPU_CACHE_LINE_SIZE - 1);
}

static u32 MmPoolHeaderSize(MEMORY_POOL *Pool) {
	return MmPoolPageStart(Pool);
}

static u32 MmPoolPageEnd(MEMORY_POOL *Pool) {
	return MmPoolHeaderSize(Pool) + (Pool->ItemsPerPage * Pool->ItemSize);
}

static void* MmPoolFirstItem(MEMORY_POOL *Pool, void* RelPoolPage) {
	if (RelPoolPage) {
		return (void*) ((u32) RelPoolPage + MmPoolPageStart(Pool));
	} else {
		return (void*) ((u32) Pool->Header + MmPoolPageStart(Pool));
	}
}

static void* MmPoolLastItem(MEMORY_POOL *Pool, void* RelPoolPage) {
	if (RelPoolPage) {
		return (void*) ((u32) RelPoolPage + MmPoolPageEnd(Pool) - Pool->ItemSize);
	} else {;
		return (void*) ((u32) Pool->Header + MmPoolPageEnd(Pool) - Pool->ItemSize);
	}
}

static MEMORY_POOL_HEADER* MmPoolHeader(void *Base) {
	return (MEMORY_POOL_HEADER*) (((u32) Base) & ~PAGEMASK);
}

static u32 MmInitializePoolPage(MEMORY_POOL *Pool, MEMORY_POOL_HEADER *Hdr) {
	if (Pool == NULL || Hdr == NULL) {
			return STATUS_INVALID_POINTER;
	}
	Hdr->Next = Pool->Header;
	Hdr->Flags = 0;
	Pool->Fragmentation++;
	if (Pool->TypeFlags & POOL_SHRINKABLE) {
		Hdr->Internal = Pool->ItemsPerPage;
	} else {
		Hdr->Internal = 0;
	}
	Pool->Header = Hdr;

	*(u32*) MmPoolLastItem(Pool, Hdr) = (u32) Pool->NextFree;
	Pool->NextFree = MmPoolFirstItem(Pool, Hdr);

	return 0;
}

/* public exported functions */

/* MmCreatePool:
 *

 * Create a memory pool with each item `PoolItemSize in size.
 * Pools should be used for high speed allocation of objects with fixed
 * size.
 *

 * Parameters:
 *
 * Pool
 *   User defined Pool Object which will be initialized by MmCreatePool.
 *   This parameter can not be NULL.
 * Data
 *   Pool data page that will be used only if POOL_STATIC is specified in
 *   `TypeFlags. `Pool.Alloc will not be called here. If POOL_STATIC is not
 *   set, this parameter will be ignored and can be NULL. `Datamust be 4 KB
 *   in size.
 * PoolItemSize
 *   Fixed size of each pool item. Valid values ranges from
 *   16 ... (4096 - SIZEOF_POOL_HEADER) Bytes.
 * TypeFlags
 *   Zero ore more of following Flags:
 *   POOL_FIXED        Pool is fixed in size. If more items are needed than
 *                     available MmAllocPoolItem call will fail.
 *   POOL_NON_PAGEABLE Pool memory will never be paged out.
 *   POOL_LOCAL        Pool memory will be created in current (caller) process
 *                     instead of kernel global shared memory. As soon as the
 *                     process terminates, pool become invalid. Only the creator
 *                     process has access to this memory pool.
 *   POOL_STATIC       Pool memory will not be created dynamically initially.
 *                     `Data parameter address points to memory that should
 *                     be used to create initial memory pool page.
 */
u32 MmCreatePool(MEMORY_POOL *Pool, void* Data, u32 PoolItemSize, MEMORY_POOL_FLAGS TypeFlags) {
	u32 Flags = 0;

	u32 Status = 0;

	if (Pool == NULL) {
			return STATUS_INVALID_POINTER;
	}


	Pool->NextFree = NULL;
	Pool->Header = NULL;
	Pool->AvlCount = 0;

	if (PoolItemSize < 16 || PoolItemSize > MAX_POOL_PAGE_SIZE - MmPoolHeaderSize(Pool)) {
			return STATUS_INVALID_RANGE;
	}

	Pool->PageSize = MAX_POOL_PAGE_SIZE;
	Pool->Size = Pool->PageSize;

	Pool->ItemSize = PoolItemSize;
	Pool->TypeFlags = TypeFlags;
	Pool->ExpandThreshold = 0;
	Pool->ShrinkThreshold = 80;
	Pool->PageCount = 1;
	Pool->MaxCount = (Pool->PageSize - MmPoolHeaderSize(Pool)) / Pool->ItemSize;
	Pool->AvlCount = Pool->MaxCount;
	Pool->ItemsPerPage = Pool->MaxCount;
	Pool->MaxSize = 0x80000000; /* 128 MB */
	Pool->Fragmentation = 0;

	if (TypeFlags & POOL_LOCAL) {
			Pool->Alloc = VmAlloc;
			Pool->Free = VmFree;
	} else {
		Pool->Alloc = GsmAlloc;
		Pool->Free = GsmFree;
	}

	if (Pool->TypeFlags & POOL_NON_PAGEABLE) {
		Flags |= PAGE_NON_PAGEABLE;
	}

	if (TypeFlags & POOL_STATIC) {

		if (Data == NULL) {
			return STATUS_INVALID_POINTER;
		}
		Pool->Header = (MEMORY_POOL_HEADER*) Data;
	} else {
		Pool->Header = NULL;
		Status = Pool->Alloc(GetPid(), (void**) &Pool->Header, Pool->PageSize, PAGE_READ_WRITE | Flags);
		if (Status) return Status;
	}

	MmInitializePoolPage(Pool, Pool->Header);
	Pool->Header->Next = NULL;

	if (TypeFlags & POOL_STATIC) {
		Pool->Header->Flags = POOL_STATIC;
	}
	Pool->NextFree = MmPoolFirstItem(Pool, NULL);
	return Status;
}

/* MmDestroyPool:
 *
 * Destroys a memory pool previously created by `MmCreatePool. Each pool page
 * will be freed automatically. One exception is the static pool page which
 * was specified in MmCreatePool as `Data with POOL_STATIC `TypeFlags.
 *

 * Parameters:
 *
 * See MmCreatePool.
 */
u32 MmDestroyPool(MEMORY_POOL *Pool) {
	if (Pool == NULL) {
			return STATUS_INVALID_POINTER;
	}


	if (Pool->Header == NULL) {
		return STATUS_CORRUPTED_DATA;
	}

	MEMORY_POOL_HEADER *p, *Hdr = Pool->Header;
	u32 PageCount = Pool->PageCount;

	while (PageCount--) {
		p = Hdr->Next;
		if ((Hdr->Flags & POOL_STATIC) == 0) {
//			printf("%s: %x\n", __FUNCTION__, Hdr);
			Pool->Free(GetPid(), (void*) Hdr, 0);
		}
		Hdr = p;
	}
}

/* MmExpandPool:
 *
 * Expands the memory pool i.e. add more pool pages so that effectively more
 * items can be stored in memory pool. New pool page will be linked into a
 * list in `MEMORY_POOL.Header. Each size of a new pool page will be 4 KB.
 *
 * Parameters:
 *
 * See MmCreatePool.
 */

u32 MmExpandPool(MEMORY_POOL *Pool) {
	u32 Status = 0;
	u32 Flags = 0;
	void* PoolPage = NULL;

	if (Pool == NULL) {
		return STATUS_INVALID_POINTER;
	}


	if (Pool->TypeFlags & POOL_FIXED) {
		return STATUS_INVALID_OPERATION;
	}

	if (Pool->Size + PAGESIZE > Pool->MaxSize) {
		return STATUS_NO_MORE_MEMORY;
	}

	if (Pool->TypeFlags & POOL_NON_PAGEABLE) {
		Flags |= PAGE_NON_PAGEABLE;
	}


	Status = Pool->Alloc(GetPid(), &PoolPage, Pool->PageSize, PAGE_READ_WRITE | Flags);
	if (Status) return Status;

	MmInitializePoolPage(Pool, PoolPage);

	Pool->PageCount++;
	Pool->Size += Pool->PageSize;
	Pool->AvlCount += Pool->ItemsPerPage;
	Pool->MaxCount += Pool->ItemsPerPage;

	return 0;
}

/* MmShrinkPool:
 *
 * Shrinks the memory pool i.e. defragment pool pages.
 *
 * Parameters:
 *
 * See MmCreatePool.
 */

u32 MmShrinkPool(MEMORY_POOL *Pool) {
	if (Pool == NULL) {
		return STATUS_INVALID_POINTER;
	}


	if (Pool->AvlCount < Pool->ItemsPerPage * 2) {
		return 0;
	}

	void *NextFree = Pool->NextFree;
	void *Next;
	MEMORY_POOL_HEADER *NextPage, *PrevPage, *FirstPage = Pool->Header;

	if ((Pool->TypeFlags & POOL_SHRINKABLE) == 0) {
		while (NextFree) {
			MmPoolHeader(NextFree)->Internal++;
			NextFree = *(void**) NextFree;
		}
	}
//	printf("Shrink in %x (%d of %d)\n", Pool->NextFree, Pool->Fragmentation, Pool->PageCount);
	NextFree = Pool->NextFree;
	Pool->NextFree = 0;
	while (NextFree) {
		Next = *(void**) NextFree;
		if (MmPoolHeader(NextFree)->Internal < Pool->ItemsPerPage) {
			*(void**) NextFree = Pool->NextFree;
			Pool->NextFree = NextFree;
		}
		NextFree = Next;
	}


	FirstPage = Pool->Header;

	PrevPage = NULL;
	while (FirstPage) {
		NextPage = FirstPage->Next;
		if (FirstPage->Internal >= Pool->ItemsPerPage && (FirstPage->Flags & POOL_STATIC) == 0) {
			if (PrevPage) {
				PrevPage->Next = NextPage;
			} else {
				Pool->Header = NextPage;
			}
			Pool->Free(GetPid(), (void*) FirstPage, 0);
			Pool->PageCount--;
			Pool->Size -= Pool->PageSize;
			Pool->AvlCount -= Pool->ItemsPerPage;
			Pool->MaxCount -= Pool->ItemsPerPage;
			Pool->Fragmentation--;
		} else {
			if ((Pool->TypeFlags & POOL_SHRINKABLE) == 0) {
				FirstPage->Internal = 0;
			}
			PrevPage = FirstPage;
		}
		FirstPage = NextPage;
	}
	return 0;
}

u32 MmGetPoolFragmentation(MEMORY_POOL* Pool, u32 *Percentage) {
	if (Pool == NULL) {
		return STATUS_INVALID_POINTER;
	}
	if (Percentage == NULL) {
		return STATUS_INVALID_POINTER;
	}
	if (Pool->NextFree == NULL) return 0;


	MEMORY_POOL_HEADER *FirstPage = Pool->Header;
	void* NextFree = Pool->NextFree;

	if ((Pool->TypeFlags & POOL_SHRINKABLE) == 0) {
		while (NextFree) {
			MmPoolHeader(NextFree)->Internal++;
			NextFree = *(void**) NextFree;
		}
	}

	*Percentage = 0;
	while (FirstPage) {
		if (FirstPage->Internal >= Pool->ItemsPerPage) (*Percentage)++;
		if ((Pool->TypeFlags & POOL_SHRINKABLE) == 0) {
			FirstPage->Internal = 0;
		}
		FirstPage = FirstPage->Next;
	}
	*Percentage = (*Percentage * 100) / Pool->PageCount;
	return 0;
}


/* MmSetPoolExpandThreshold:
 *
 * Sets the Threshold afterward `MmExpandPool will be called. This is initially
 * set to 1, i.e. pool memory will be expanded only if only one more free item
 * is available, otherwise if `MEMORY_POOL.AvlCount <= `MEMORY_POOL.Threshold
 * `MmExpandPool will be called. This is only usefull in critical situations
 * especially if combined with POOL_STATIC in system initialization for example.
 *
 * Parameters:
 *
 * See MmCreatePool.
 * Threshold
 *   Threshold value. This can not be more or equal to `MEMORY_POOL.MaxCount.
 */

u32 MmSetPoolExpandThreshold(MEMORY_POOL *Pool, u32 Threshold) {
	if (Pool == NULL) {
			return STATUS_INVALID_POINTER;
	}
	if (Threshold >= Pool->MaxCount) {
		return STATUS_INVALID_ARGUMENT;
	}
	if (Threshold >= Pool->ExpandThreshold) {
		if ((Pool->TypeFlags & POOL_FIXED) == 0) {
			MmExpandPool(Pool);
		}
	}
	Pool->ExpandThreshold = Threshold;
	return 0;
}

/* MmSetPoolShrinkThreshold:
 *
 * Sets the Threshold afterward `MmShrinkPool will be called. This is initially
 * set to 50%, i.e. pool memory will be shrinked only if at least 50% of pool
 * pages are unused.
 *
 * Parameters:
 *
 * See MmCreatePool.
 * Threshold
 *   Threshold value. Valid values are 0 ... 100.
 */

u32 MmSetPoolShrinkThreshold(MEMORY_POOL *Pool, u32 Threshold) {
	if (Pool == NULL) {
			return STATUS_INVALID_POINTER;
	}
	if (Threshold > 100) {
		return STATUS_INVALID_ARGUMENT;
	}
	if (Pool->TypeFlags & POOL_SHRINKABLE) {
		if (Threshold >= Pool->ShrinkThreshold) {
			if ((Pool->Fragmentation * 100) / Pool->PageCount >= Threshold) {
				MmShrinkPool(Pool);
			}
		}
	}
	Pool->ShrinkThreshold = Threshold;
	return 0;
}


/* MmSetPoolExpandSize:
 *
 * Sets the size a memory pool can expand dynamically if no more free items
 * are available. If pool is currently larger than `Size memory pool will
 * not be shrinked!
 *
 * Parameters:
 *
 * Size
 *   Size a memory pool can expand. This could not be less than 4 KB (PAGESIZE).
 */

u32 MmSetPoolExpandSize(MEMORY_POOL *Pool, u32 Size) {
	if (Size < PAGESIZE) return STATUS_INVALID_ARGUMENT;
	Pool->MaxSize = Size;
	return 0;
}

u32 MmAllocPoolItem(MEMORY_POOL *Pool, void** Data) {
	u32* NextFree;
	u32 Status = 0;

	if (Pool == NULL || Data == NULL) {
			return STATUS_INVALID_POINTER;
	}

	if (Pool->AvlCount <= Pool->ExpandThreshold) {
		if (Pool->TypeFlags & POOL_FIXED) {
			return STATUS_NO_MORE_MEMORY;
		} else {
			Status = MmExpandPool(Pool);
			if (Status) return Status;
		}
	}

	Pool->AvlCount--;
	NextFree = Pool->NextFree;
	Pool->NextFree = (void*) *NextFree;

	if (Pool->NextFree == NULL) {
		Pool->NextFree = (void*) ((u32) NextFree + Pool->ItemSize);
	}

	*Data = (void*) NextFree;

/*	if (Pool->TypeFlags & POOL_SHRINKABLE) {
		if (MmPoolHeader(*Data)->Internal == Pool->ItemsPerPage) {
			MmPoolHeader(*Data)->Internal--;
			Pool->Fragmentation--;
		}
	}*/
	return 0;
}

u32 MmFreePoolItem(MEMORY_POOL *Pool, void* Data) {
	if (Pool == NULL || Data == NULL) {
			return STATUS_INVALID_POINTER;
	}

	*(void**) Data = Pool->NextFree;
	Pool->NextFree = Data;
	Pool->AvlCount++;
/*	if (Pool->TypeFlags & POOL_SHRINKABLE) {
		if (++MmPoolHeader(Data)->Internal == Pool->ItemsPerPage) {
			Pool->Fragmentation++;
			u32 Frag = (Pool->Fragmentation * 100) / Pool->PageCount;
			if (Frag >= Pool->ShrinkThreshold) {
				MmShrinkPool(Pool);
			}

		}
	}*/
	return 0;
}


char* StrFormatSize(u32 Value, char *Data) {
	char ta[16], tb[16];

	u32 a, b;
	if (Value < 1000) {
		sprintf(Data, "%u Byte", Value);
	} else if (Value < (1000 * 1000)) {
		a = Value / 1024;
		b = Value % 1024;
		sprintf(ta, "%u", a);
		sprintf(tb, "%u", b);
		tb[2] = '\0';
		sprintf(Data, "%s.%s KB", ta, tb);
	} else if (Value < (1000 * 1000 * 1000)) {
		a = Value / (1024 * 1024);
		b = Value % (1024 * 1024);
		sprintf(ta, "%u", a);
		sprintf(tb, "%u", b);
		tb[2] = '\0';
		sprintf(Data, "%s.%s MB", ta, tb);
	} else {
		a = Value / (1024 * 1024 * 1024);
		b = Value % (1024 * 1024 * 1024);
		sprintf(ta, "%u", a);
		sprintf(tb, "%u", b);
		tb[2] = '\0';
		sprintf(Data, "%s.%s GB", ta, tb);
	}
	return Data;
}
/*
int main() {
	MEMORY_POOL Pool;
	char Temp[16];
	void *Data[2000000];
	unsigned int i;

	MmCreatePool(&Pool, NULL, 200, 0);
	for (i = 0; i < sizeof(Data) / sizeof(void*); i++) {
		Data[i] = NULL;
		MmAllocPoolItem(&Pool, (void**) &Data[i]);
		if ((i & 0xff)==0) {
			StrFormatSize(Pool.Size, Temp);
			printf("<A> Data = %x, (Avl = %d, NF = %x, Sz = %s (I = %x)\n", Data[i], Pool.AvlCount, Pool.NextFree, Temp, i);
		}
		if (Data[i] == NULL) {
			printf("winerror=%u\n", GetLastError());
			break;
		}
	}
	for (i = 0; i < sizeof(Data) / sizeof(void*); i++) {
		if (Data[i] == NULL) break;
		if (i & (rand() % 41171)) {
			MmFreePoolItem(&Pool, Data[i]);
			Data[i] = 0;
		}
		if ((i & 0xff)==0) {
			StrFormatSize(Pool.Size, Temp);
			printf("<F> Data = %x, (Avl = %d, NF = %x, Sz = %s (I = %x)\n", Data[i], Pool.AvlCount, Pool.NextFree, Temp, i);
		}
	}
	printf("Before Shrink: %d pages (%d free items)\n", Pool.PageCount, Pool.AvlCount);

	MmShrinkPool(&Pool);

	printf("After Shrink: %d pages (%d free items)\n", Pool.PageCount, Pool.AvlCount);
	MmDestroyPool(&Pool);
	getchar();
}
*/
