#include <stdlib.h>
#include <stdio.h>
#include "../general.h"
#include "../pool/pool.h"
#include "bst.h"

TREE *TreeCreate(u32 NodeSize, void *PoolBase, u32 Flags) {
	MEMORY_POOL Pool = {0};
	TREE *Tree = NULL;

	if (MmCreatePool(&Pool, PoolBase, NodeSize, Flags) == STATUS_SUCCESS) {
		MmAllocPoolItem(&Pool, (void*) &Tree);
		/* Allocate so many Items that entire allocation matches TREE size */
		u32 PoolItems = (sizeof(TREE) + NodeSize - 1) / NodeSize - 1;
		while (PoolItems--) {
			void *Temp;
			MmAllocPoolItem(&Pool, &Temp);
		}
		memcpy(&Tree->Pool, &Pool, sizeof(Pool));
	}
	return Tree;
}

void TreeDestroy(TREE *Tree) {
	MmDestroyPool(&Tree->Pool);
}

TREE_NODE *TreeCreateNode(TREE *Tree, void *Key) {
	TREE_NODE *Node = NULL;
	if (MmAllocPoolItem(&Tree->Pool, (void*) &Node) == STATUS_SUCCESS) {
		Node->Key = Key;
		Node->Parent = NULL;
		Node->Branch[TREE_LEFT] = NULL;
		Node->Branch[TREE_RIGHT] = NULL;
	}
	return Node;
}

void TreeFreeNode(TREE *Tree, TREE_NODE *Node) {
	MmFreePoolItem(&Tree->Pool, Node);
}


/* Minimum Item of a binary tree */

TREE_NODE *TreeMin(TREE_NODE *Node) {
	while (Node->Branch[TREE_LEFT]) {
		Node = Node->Branch[TREE_LEFT];
	}
	return Node;
}

/* Maximum Item of a binary tree */

TREE_NODE *TreeMax(TREE_NODE *Node) {
	while (Node->Branch[TREE_RIGHT]) {
		Node = Node->Branch[TREE_RIGHT];
	}
	return Node;
}

/* Find successor item */

TREE_NODE *TreeSucc(TREE_NODE *Node) {
	TREE_NODE *Pred = Node->Parent;
	if (Node->Branch[TREE_RIGHT]) return TreeMin(Node->Branch[TREE_RIGHT]);
	while (Pred && Pred->Key <= Node->Key) {
		Pred = Pred->Parent;
	}
	return Pred;
}

TREE_NODE *TreeSearchKey(TREE_NODE *Parent, void *Key) {
	if (Parent == NULL) return NULL;
	if ((u32) Key < (u32) Parent->Key) {
		if (Parent->Branch[TREE_LEFT]) {
			return TreeSearchKey(Parent->Branch[TREE_LEFT], Key);
		}
	} else {
		if (Parent->Branch[TREE_RIGHT]) {
			return TreeSearchKey(Parent->Branch[TREE_RIGHT], Key);
		}
	}
	return Parent;
}


TREE_NODE *TreeInsertKey(TREE *Tree, void *Key) {
	TREE_NODE *Node = TreeCreateNode(Tree, Key);
	if (Node == NULL) return NULL;

	TREE_NODE *KeyNode = TreeSearchKey(Tree->Root, Key);

	Node->Parent = KeyNode;
	if (KeyNode) {
		if ((u32) Key < (u32) KeyNode->Key) {
			Node->Branch[TREE_RIGHT] = KeyNode->Branch[TREE_LEFT];
			KeyNode->Branch[TREE_LEFT] = Node;
			if (Node->Branch[TREE_RIGHT]) Node->Branch[TREE_RIGHT]->Parent = Node;
		} else {
			Node->Branch[TREE_LEFT] = KeyNode->Branch[TREE_RIGHT];
			KeyNode->Branch[TREE_RIGHT] = Node;
			if (Node->Branch[TREE_LEFT]) Node->Branch[TREE_LEFT]->Parent = Node;
		}
	} else {
		Tree->Root = Node;
	}
	return Node;
}

void TreeRemoveKey(TREE *Tree, void* Key) {
	TREE_NODE *KeyNode = TreeSearchKey(Tree->Root, Key);
	TREE_NODE *ToDelete, *Son, *Parent;

	if (KeyNode && KeyNode->Key == Key) {
		if (KeyNode->Branch[TREE_LEFT] == 0 || KeyNode->Branch[TREE_RIGHT] == 0)
			ToDelete = KeyNode;
		else
			ToDelete = TreeSucc(KeyNode);

		if (ToDelete->Branch[TREE_LEFT])
			Son = ToDelete->Branch[TREE_LEFT];
		else
			Son = ToDelete->Branch[TREE_RIGHT];

		Parent = ToDelete->Parent;

		if (Son) Son->Parent = Parent;

		if (Parent == NULL) {
			Tree->Root = Son;
		} else {
			if (ToDelete == Parent->Branch[TREE_LEFT])
				Parent->Branch[TREE_LEFT] = Son;
			else
				Parent->Branch[TREE_RIGHT] = Son;
		}

		if (ToDelete != KeyNode) {
			KeyNode->Key = ToDelete->Key;
		}

		TreeFreeNode(Tree, ToDelete);
	}
}

int main(int argc, char* argv[]) {
	TREE *Tree = TreeCreate(sizeof(TREE_NODE), NULL, 0);
	int i;
	srand(1000);
	for (i = 0; i < 1000000; i++) {
		int n = rand();
		if ((i & 4095) == 4095) printf("Insert: %d\n", n);
		TreeInsertKey(Tree, (void*) n);
	}
	printf("pool pages: %d, size: %d\n", Tree->Pool.PageCount, Tree->Pool.Size);
	srand(1000);
	for (i = 0; i < 1000000; i++) {
		TreeRemoveKey(Tree, (void*) i);
	}
	TreeDestroy(Tree);
	return 0;
}
